// CStorage.cpp
//**********************************************************************************************
//
//
//				(C) Copyright 1998 by Dipl. Phys. Joerg Plewe, HARDCODE Development
//					All rights reserved. Copying, modification,
//					distribution or publication without the prior written 
//					consent of the author is prohibited.
//
//
//	Filename:		CStorage.cpp
//	Author:			Herkules
//  Created:		9810xx 
//
//	Module:			Storage
//  Dependecies:	Files in directories Memory and Library
//
//  History:
//  $Date$	$Who$	$What$
//
/*	Purpose:

Implement the exported classes as far as possible. Objects implemented here will be used by
the user of the DLL.
Relies on implementations of the abstract classes within the 'drivers' Memory and Library

*/
//**********************************************************************************************



#define EXPORT_STORAGE




#include "CStorage.h"
#include "Memory/CStorageMemory.h"
#include "Library/CStorageLibrary.h"



#include <windows.h>
#include <assert.h>


#include <fstream>




//-----------------------------------------------------------------------------CreatePOP-----------
// Herkules
// Global function: create a new, unique POP
//

POP CreatePOP() 
{
	POP Pop;


	LARGE_INTEGER LowerPart; 
	if( QueryPerformanceCounter(&LowerPart) ) {
		Pop = LowerPart.LowPart;
	} else {
		Pop = GetTickCount();
	}
	
	SYSTEMTIME Time;
	GetSystemTime(&Time);
    
	__int64 n =	Time.wYear; 
							n *= Time.wMonth; 
    if(Time.wDayOfWeek)		n *= Time.wDayOfWeek; 
    if(Time.wDay)			n *= Time.wDay; 
    if(Time.wHour)			n *= Time.wHour; 
    if(Time.wMinute)		n *= Time.wMinute; 
    if(Time.wSecond)		n *= Time.wSecond; 
    if(Time.wMilliseconds)	n *= Time.wMilliseconds; 

	// xor higher and lower int together
	n ^= n >> 32;

	// put lower int to high int  
	n = n << 32;

	// and use it as high int for pop
	Pop |= n;


	return Pop;

} // CreatePOP





//-------------------------------------------------------------------------------------------------
//
//
// Implementation of CStoredClass                                              CStoredClass
//
//
//-------------------------------------------------------------------------------------------------

//-----------------------------------------------------------------------------CStoredClass--------
// Herkules
// Constructor
//

CStoredClass::CStoredClass( const string _Classname, const STOREDCLASS_ID _ClassID, IStorageEngine* _pStorage, IStoredClass* const _pBaseClass ) :
IStoredClass( _Classname, _ClassID, _pStorage, _pBaseClass ),
m_LastVarID( 0 )
{

	//
	// If there is a base class, m_LastVarID will never be used.
	//

	// nothing to do here

}



//-----------------------------------------------------------------------------~CStoredClass-------
// Herkules
// Destructor
//

CStoredClass::~CStoredClass() 
{


	std::list<CStoredVar*>::iterator It;

	for( It = m_lVars.begin(); It != m_lVars.end(); ++It )
	{
		delete (*It);
	}
	
	
	m_lVars.clear();

}





//-----------------------------------------------------------------------------CreateVar-----------
// Herkules
// protected
// create a new variable add it to the list
//

CStoredVar* 
CStoredClass::CreateVar( const string& _Varname, STOREDVAR_TYPE _Type, STOREDVAR_ID _ID )
{


	// first, create a new instance
	CStoredVar* pStoredVar = new CStoredVar( _Varname, _Type, _ID );

	// done
	return pStoredVar;

} // CreateVar




//-----------------------------------------------------------------------------NewVarID------------
// Herkules
// Create a new variable ID
//

int 
CStoredClass::NewVarID()
{

	if( m_pBaseClass )
		return static_cast<CStoredClass*>(m_pBaseClass)->NewVarID();
	else
		return m_LastVarID++;


} // NewVarID




//-----------------------------------------------------------------------------AddVar--------------
// Herkules
// Add a variable for this class to repository
//

IStoredVar* 
CStoredClass::AddVar( const string& _Varname, STOREDVAR_TYPE _Type )
{
	
	IStoredVar* pVar = NULL;
	
	//
	// Try to find a Var like this in the repository. If it also fits in type
	// we can use it. If the type does not fit, this is an error!
	//
	if( ( pVar = GetVar( _Varname ) ) ) 
	{
		
		if( pVar->GetType() == _Type )
			return pVar;

		// Type does not fit: ERROR
		return NULL;

	}
	else
	{
		
		// Use a virtual function to create the variable
		pVar = CreateVar( _Varname, _Type, NewVarID() );

		// ... and add it to the repository
		AddVarToRepository( static_cast<CStoredVar*>( pVar ) );

		return pVar;
	}


} // AddVar






//-----------------------------------------------------------------------------GetVar--------------
// Herkules
// Get a variable from its name
//

IStoredVar* 
CStoredClass::GetVar( const string& _Varname )
{

	//
	// Check our own list for the variable FIRST
	//
	
	std::list<CStoredVar*>::iterator It;

	for( It = m_lVars.begin(); It != m_lVars.end(); ++It )
	{
		if( (*It)->Name() == _Varname )
			return *It;
	}


	// 
	// Then ask the base class for the variable
	//
	if( GetBaseClass() )
	{
		IStoredVar* Var = GetBaseClass()->GetVar( _Varname );
		
		// If the base class has the variable, return it.
		if( Var )
			return Var;
	}




	// Nothing found
	return NULL;

} // GetVar




//-----------------------------------------------------------------------------AddVarToRepository--
// Herkules
// add an existing variable to the repository
//

void 
CStoredClass::AddVarToRepository( CStoredVar* _pVar )
{

	// adjust internal variable counter
	if( _pVar->GetID() > m_LastVarID )	m_LastVarID = _pVar->GetID();

	//m_lVars.insert( m_lVars.end(), pStoredVar );
	m_lVars.push_back( _pVar );

} // AddVarToRepository







// ----------------------------------------------------------------------------
// fire an event that the content of the class has changed
//

void 
CStoredClass::FireChangedEvent( IStoredObject* _pObject, IStoredVar* _pVar )
{

	GetStorage()->FireChangedEvent( this, _pObject, _pVar );

}





//-------------------------------------------------------------------------------------------------
//
//
// Implementation of CStoredObject                                             CStoredObject
//
//
//-------------------------------------------------------------------------------------------------




//-----------------------------------------------------------------------------FireChangedEvent----
// Herkules
// Fire an event that the content of the object has changed
//

void 
CStoredObject::FireChangedEvent( IStoredVar* _pVar )
{

	GetClass()->FireChangedEvent( this, _pVar );

} // FireChangedEvent











//-------------------------------------------------------------------------------------------------
//
//
// Implementation of CStorage                                                  CStorage
//
//
//-------------------------------------------------------------------------------------------------

//-----------------------------------------------------------------------------CStorage------------
// Herkules
// Constructor
//

CStorage::CStorage( IStorageEvent* _pEvent ) :
m_pStorageEvent( _pEvent ),
m_StorageImplementation( 0 )
{

}





// ----------------------------------------------------------------------------
//

IStorageEngine* 
IStorageEngine::New( int _StorageType, IStorageEvent* _pEvent )
{

	IStorageEngine* pNew = NULL;

	switch( _StorageType )
	{
	case IStorageEngine::Memory:
		{
			pNew =  new CStorageMemory( _pEvent );
		} 
		break;
	case IStorageEngine::Library:
		{
			pNew =  new CStorageLibrary( _pEvent );
		} 
		break;

	}

//	pNew->m_StorageImplementation = _StorageType;

	return pNew;

} // Constructor




//-----------------------------------------------------------------------------
// Destructor
//

CStorage::~CStorage( )
{
	std::vector<IStoredClass*>::iterator It;

	for( It = m_lClasses.begin(); It != m_lClasses.end(); ++It )
	{
		delete (CStoredClass*)(*It);
	}
	
	m_lClasses.clear();

} // Destructor



#define FOREACHSTOREDTYPE( TYPE, SRC, CURRENT )					\
{																\
	IStored##TYPE##List *pTypeList = (SRC)->Get##TYPE##List();	\
	for( IStored##TYPE *CURRENT = pTypeList->First();			\
				CURRENT != NULL;								\
				CURRENT = pTypeList->Next() )					\
	{


#define STOREDTYPEENDFOR										\
	}															\
	pTypeList->Release();										\
}






//------------------------------------------------------------------------------
// Herkules
// merge with the repository from another database
//
bool 
CStorage::MergeRepository( const CStorage& _Source )
{

	FOREACHSTOREDTYPE( Class, &_Source, pSourceClass )
	{	

		// create a new class with the sources name
		// if it fails - don't care. Means the class is already present
		IStoredClass* pDestClass = AddClass( pSourceClass->Name() );
		if( ! pDestClass )
			pDestClass = GetClass( pSourceClass->Name() );

		// if there still is no class pointer, something is wrong!
		assert( pDestClass );

		
		//
		// no loop on the variables
		//
		FOREACHSTOREDTYPE( Var, pSourceClass, pSourceVar )
		{
			IStoredVar* pDestVar = pDestClass->AddVar( pSourceVar->Name(), pSourceVar->GetType() );
			if( ! pDestVar ) 
			{
				pDestVar = pDestClass->GetVar( pSourceVar->Name() );

				// source and dest must be of same type!
				if( pDestVar->GetType() != pSourceVar->GetType() )
					return false;
			}
			assert( pDestVar );
		
		} STOREDTYPEENDFOR
	} STOREDTYPEENDFOR

	return true;

} // CopyRepository





//------------------------------------------------------------------------------
// Herkules
// helper template function: 
//
//

template <class T> static bool CopySingleScalarValue( 
								IStoredObject* _pSrcObj, IStoredVar* _pSrcVar, 
								IStoredObject* _pDstObj, IStoredVar* _pDstVar, T& _Value 
								)
{
	// try to get the value ...
	if( _pSrcObj->GetValue( _pSrcVar, _Value ) )

		// and add it to the destination 
		return _pDstObj->AddValue( _pDstVar, _Value );

	// could not find value
	return false;

} // CopySingleScalarValue



//------------------------------------------------------------------------------
// Herkules
// helper template function: 
// Copy a single, array-typed value from one database to another
// If the array is large (> _FIXEDBUFFERSIZE) allocate memory dynamically 
//

#define _FIXEDBUFFERSIZE 1024
 
template <class T> static bool CopySingleArrayValue( 
								IStoredObject* _pSrcObj, IStoredVar* _pSrcVar, 
								IStoredObject* _pDstObj, IStoredVar* _pDstVar, T& _Value 
								)
{

	//
	// locals
	//
	int Dim = 0;
	bool ret;
	T FixedBuffer[ _FIXEDBUFFERSIZE ];
	T* pBuffer = FixedBuffer;


	// get dimension of array
	// no data is transfered bc. Dim=0
	_pSrcObj->GetValue( _pSrcVar, pBuffer, Dim );

	if( Dim >= _FIXEDBUFFERSIZE )
		// allocate new array
		pBuffer = new T[ Dim ];


	// try to read the buffer
	if( _pSrcObj->GetValue( _pSrcVar, pBuffer, Dim ) )

		// and add it the destination
		ret = _pDstObj->AddValue( _pDstVar, pBuffer, Dim );


	if( Dim >= _FIXEDBUFFERSIZE )
		delete[] pBuffer;

	return ret;

} // CopySingleArrayValue



//------------------------------------------------------------------------------
// Herkules
// helper function: Copy a single value from one database to the other
//

static bool CopyStoredValue( IStoredObject* _pSrcObj, IStoredVar* _pSrcVar, 
								IStoredObject* _pDstObj, IStoredVar* _pDstVar )
{

	switch( _pSrcVar->GetType() )
	{
	case IStoredVar::Byte:
		{
			unsigned char Value;
			return CopySingleScalarValue( _pSrcObj, _pSrcVar, _pDstObj, _pDstVar, Value ); 
		}
	case IStoredVar::Short:
		{
			short Value;
			return CopySingleScalarValue( _pSrcObj, _pSrcVar, _pDstObj, _pDstVar, Value ); 
		}
	case IStoredVar::Int:
		{
			int Value;
			return CopySingleScalarValue( _pSrcObj, _pSrcVar, _pDstObj, _pDstVar, Value ); 
		}
	case IStoredVar::Float:
		{
			float Value;
			return CopySingleScalarValue( _pSrcObj, _pSrcVar, _pDstObj, _pDstVar, Value ); 
		}
	case IStoredVar::Double:
		{
			double Value;
			return CopySingleScalarValue( _pSrcObj, _pSrcVar, _pDstObj, _pDstVar, Value ); 
		}
	case IStoredVar::String:
		{
			string Value;
			return CopySingleScalarValue( _pSrcObj, _pSrcVar, _pDstObj, _pDstVar, Value ); 
		}

		//
		// arrays
		//


	case IStoredVar::ByteArray:
		{
			unsigned char Value;
			return CopySingleArrayValue( _pSrcObj, _pSrcVar, _pDstObj, _pDstVar, Value ); 
		}
	case IStoredVar::ShortArray:
		{
			short Value;
			return CopySingleArrayValue( _pSrcObj, _pSrcVar, _pDstObj, _pDstVar, Value ); 
		}
	case IStoredVar::IntArray:
		{
			int Value;
			return CopySingleArrayValue( _pSrcObj, _pSrcVar, _pDstObj, _pDstVar, Value ); 
		}
	case IStoredVar::FloatArray:
		{
			float Value;
			return CopySingleArrayValue( _pSrcObj, _pSrcVar, _pDstObj, _pDstVar, Value ); 
		}
	case IStoredVar::DoubleArray:
		{
			double Value;
			return CopySingleArrayValue( _pSrcObj, _pSrcVar, _pDstObj, _pDstVar, Value ); 
		}

	}

	return false;

} // CopyStoredValue





//------------------------------------------------------------------------------
// Herkules
// copy the content from another database
// This function does not change the repository of the destination database
// Only classes and variables
//

bool 
CStorage::CopyContent( const CStorage& _Source )
{


	//
	// iterate on all classes in the destination database
	//
	FOREACHSTOREDTYPE( Class, this, pDestClass )
	{	

		//
		// iterate all objects 
		//
		
		// first, get the same class in the source  
		IStoredClass* pSrcClass = _Source.GetClass( pDestClass->Name() );

		// if it exists
		if( pSrcClass ) 
		{

			// 
			// iterate on all object in source class
			//
			FOREACHSTOREDTYPE( Object, pSrcClass, pSrcObj )
			{
				// Create a new object in destination
				IStoredObject* pNewObj = pDestClass->AddObject();

				// 
				// iterate in all variables in destination class
				//
				FOREACHSTOREDTYPE( Var, pDestClass, pVar )
				{
					// search destination variable in source
					IStoredVar* pSrcVar = pSrcClass->GetVar( pVar->Name() );

					// if the variable exists in source also, copy data
					if( pSrcVar && pSrcObj->HasValue( pSrcVar ) )

						// try to copy the value
						if( ! CopyStoredValue( pSrcObj, pSrcVar, pNewObj, pVar ) )
							return false;

				} STOREDTYPEENDFOR

				// save new object
				pNewObj->Flush( false );

				// throw it away
				pNewObj->Release();

				// release also the source!
				pSrcObj->Release();

			} STOREDTYPEENDFOR

		}


	} STOREDTYPEENDFOR


	return true;

} // CopyContent






//-----------------------------------------------------------------------------Close---------------
// Herkules
// Close the lib
//

bool 
CStorage::Close() 
{
	return true;

} // Close








//-----------------------------------------------------------------------------FireChangedEvent----
// Herkules
// Fire an event that the library changed content
//

void 
CStorage::FireChangedEvent( IStoredClass* _pClass, IStoredObject* _pObject, IStoredVar* _pVar )
{

	// fire if an CStorageEvent was set
	if( m_pStorageEvent )
		m_pStorageEvent->Fire( _pClass, _pObject, _pVar );

} // FireChangedEvent




//-----------------------------------------------------------------------------AddClass------------
// Herkules
// Add a class to to the repository
//

IStoredClass* 
CStorage::AddClass( const string& _Classname, IStoredClass* const _pBaseClass )
{
	CStoredClass* pStoredClass = NULL;

	//
	// First check wether this class already IS in repository
	// If it is, take it home!
	//
	if( ! ( pStoredClass = FindClassInRepository( _Classname ) ) ) 
	{

		// Create a new classes instance with this virtual creator method
		pStoredClass = CreateClass( _Classname, m_lClasses.size(), _pBaseClass );

		// Add the new guy to our repository
		AddClassToRepository( pStoredClass );

	}


	// return either NULL or the new object or an existing class
	return pStoredClass;

} // AddClass





//-----------------------------------------------------------------------------GetClass------------
// Herkules
// Find a class from its name
//

IStoredClass* 
CStorage::GetClass( const string& _Classname ) const 
{
	return FindClassInRepository( _Classname );

} // GetClass



//-----------------------------------------------------------------------------GetClass------------
// Herkules
// Find a class from its ID
//

IStoredClass* 
CStorage::GetClass( const STOREDCLASS_ID _ClassID ) const 
{

	assert( m_lClasses[ _ClassID ]->GetID() == _ClassID );

	return m_lClasses[ _ClassID ];

} // GetClass





//-----------------------------------------------------------------------------AddClassToRepository
// Herkules
// Add an existing calls to the repository
//

void 
CStorage::AddClassToRepository( CStoredClass* _pNewClass )
{

	// just add the class to the list
	m_lClasses.push_back( _pNewClass );

} // AddClassToRepository





// ----------------------------------------------------------------------------
//

CStoredClass* 
CStorage::FindClassInRepository( const string& _Classname ) const 
{

	vector<IStoredClass*>::const_iterator It;

	//
	// search in all class for an instance with the name given
	//
	for( It = m_lClasses.begin(); It != m_lClasses.end(); ++It )
	{
		if( (*It)->Name() == _Classname )
			return (CStoredClass*)*It;
	}

	return NULL;

}





// ----------------------------------------------------------------------------
//
// parse and interpret an ascii file in specified format
//
//

#define LINEBUFFERSIZE 65536

struct {
	const char* Name;
	STOREDVAR_TYPE Type;
} TypeStringLookup[] = {
	{ "Byte",	IStoredVar::Byte	},
	{ "Short",	IStoredVar::Short	},
	{ "Int",	IStoredVar::Int	},
	{ "LongInt",IStoredVar::LongInt},
	{ "Float",	IStoredVar::Float	},
	{ "Double",	IStoredVar::Double },
	{ "String",	IStoredVar::String },
	{ "Reference",	IStoredVar::Reference },
	{ "", 0 }
};




template <class T> static void _ReadTypedArray( FILE* _f, IStoredObject* _pObject, IStoredVar* _pVar, int _ArraySize, char* _pFmt, T _Init )
{
	int i;
	T* pData = new T[_ArraySize];
	unsigned char Buffer[10];

	for( i = 0; i < _ArraySize; i++ )
	{
		pData[i] = _Init;						
		fscanf( _f, _pFmt, Buffer );

		pData[i] = *reinterpret_cast<T*>(Buffer);
	}

	_pObject->AddValue( _pVar, pData, _ArraySize );
	delete[] pData;
}




//------------------------------------------------------------------------------------------------
// read a special ini file and create database content
//

bool 
CStorage::ReadTextFile( const char* _Filename )
{

	//
	// locals
	//
	int i;
	FILE* f;
	char buf[LINEBUFFERSIZE];
	int ItemsRead;


	// initialize the pointer the the CStoredObject to be created
	IStoredObject* pObject = NULL;


	// open the file
	f = fopen( _Filename, "r" );
	if( !f ) return false;



	// read first class specifier
	ItemsRead = fscanf( f, "%s", buf );
	
	while( ItemsRead != EOF )
	{
		// check and strip off []

		// the first char has to be '[', the last has to be ']'
		if( buf[0] != '[' || buf[ strlen(buf) -1 ] != ']' )	goto errorexit;

		// replace ']' by '\0'
		buf[ strlen(buf) -1 ] = '\0';

		
		//
		// get the class or create a new one
		// at buf+1 the name begins, behind '['
		IStoredClass* pClass = GetClass( buf+1 );
		if( ! pClass ) pClass = AddClass( buf+1 );



		//
		// now create a new object for that class
		//

		// is there already is an object from the last turn, flush and realese it
		if( pObject ) 
		{
			pObject->Flush();
			pObject->Release();
		}

		// create the new object
		pObject = pClass->AddObject();


		//
		// get the values
		//

		// read the next string. Should be of kind <Type>:<Name>[<Dimension>]
		ItemsRead = fscanf( f, "%s", buf );

		// we do this until we reach the end of file or the next class 
		// department beginning with '[' again
		while( ItemsRead != EOF  &&  buf[0] != '[' )
		{
			char* Name;
			STOREDVAR_TYPE Type = 0;
			int ArraySize = 0;

			
			// string now contains TYPE:NAME[DIM]
			// [DIM] is optional

			// parse the string for the ':', dividung type from name
			Name = strtok( buf, ":" );

			// now dispatch for all know types
			for( i = 0; i < IStoredVar::NumberOfScalarTypes; i++ )
			{

				// compare typename
				if( stricmp( buf, TypeStringLookup[i].Name ) == 0 )
				{
					// remenber what type it is 
					Type = TypeStringLookup[i].Type;

					// leave the loop
					break;
				}
			}

			// if we did not find anything, exit 
			if( ! Type )	goto errorexit;


			//
			// otherwise, get or create a variable with correct type
			//

			// parse the string for the next whitespace delimiter,
			// either ' ' or '\t'
			Name = strtok( NULL, " \t" );



			//
			// check wether an array is needed
			//
			if( Name[strlen(Name)-1] == ']' )		
			{
				char* DimString;

				// start a new strtok sequence
				Name = strtok( Name, "[" );
		
				// get the substring with the dimension
				DimString = strtok( NULL, "]" );

				// convert it to number
				ArraySize = atoi( DimString );
			}


			//
			// is ArraySize is > 0, we need a array
			//

			// set the ARRAY flag
			if( ArraySize ) Type |= VARTYPEARRAYMASK;


			// try to locate the variable
			IStoredVar* pVar = pClass->GetVar( Name );

			// if the variable does not exist, create a new one
			if( ! pVar )	
				pVar = pClass->AddVar( Name, Type );
		

			//
			// now that we have a variable, set the value
			// 
			switch( Type )
			{
			case IStoredVar::Byte:
				{
					Byte Value;
					ItemsRead = fscanf( f, "%d", &Value );
					pObject->AddValue( pVar, Value );
				}
				break;

			case IStoredVar::Short:
				{
					Short Value;
					ItemsRead = fscanf( f, "%d", &Value );
					pObject->AddValue( pVar, Value );
				}
				break;
			
			case IStoredVar::Int:
				{
					int Value;
					ItemsRead = fscanf( f, "%d", &Value );
					pObject->AddValue( pVar, Value );
				}
				break;

			case IStoredVar::Float:
				{
					float Value;
					ItemsRead = fscanf( f, "%f", &Value );
					pObject->AddValue( pVar, Value );
				}
				break;


			case IStoredVar::Double:
				{
					double Value;
					ItemsRead = fscanf( f, "%lf", &Value );
					pObject->AddValue( pVar, (double)Value );
				}
				break;


			case IStoredVar::String:
				{
					int len;
					int c;
					// skip one following whitespace character
					c = fgetc( f );

					ZeroMemory( buf, LINEBUFFERSIZE );
					ItemsRead = (int)fgets( buf, LINEBUFFERSIZE, f  );

					// remove trailing line delimiters
					len = strlen( buf );
					while( buf[len-1] == '\r' || buf[len-1] == '\n' )
					{
						len--;
						buf[len] = '\0';
					}

					string str = buf;
					pObject->AddValue( pVar, str );
				}
				break;

			case IStoredVar::ByteArray:
				_ReadTypedArray<unsigned char>( f, pObject, pVar, ArraySize, "%hd", (unsigned char)0 );
				break;				

			case IStoredVar::ShortArray:
				_ReadTypedArray<short>( f, pObject, pVar, ArraySize, "%hd", (short)0 );
				break;				

			case IStoredVar::IntArray:
				_ReadTypedArray<int>( f, pObject, pVar, ArraySize, "%d", (int)0 );
				break;				

			case IStoredVar::FloatArray:
				_ReadTypedArray<float>( f, pObject, pVar, ArraySize, "%f", (float)0.0f );
				break;				

			case IStoredVar::DoubleArray:
				_ReadTypedArray<double>( f, pObject, pVar, ArraySize, "%lf", (double)0.0 );
				break;				
			
			}


			
			//
			// get next line
			//
			ItemsRead = fscanf( f, "%s", buf );


		}

	}



	//
	// flush the last object
	// 
	if( pObject ) 
	{
		pObject->Flush();
		pObject->Release();
	}




	fclose( f );
	return true;

errorexit:
	fclose( f );
	return false;

} // ReadTextFile






//------------------------------------------------------------------------------------------------
// write a special ini file from database content
//

bool 
CStorage::WriteTextFile( const char* _Filename )
{

	//
	// locals
	//

	// create the output file
	ofstream File( _Filename ); 

	
	FOREACHSTOREDTYPE( Class, this, pClass )
	{	

		// 
		// iterate on all object in source class
		//
		FOREACHSTOREDTYPE( Object, pClass, pObj )
		{

			// write a new object header [classname]
			File << "[" << pClass->Name() << "]" << endl;


			FOREACHSTOREDTYPE( Var, pClass, pVar )
			{

				// if the object has that variable
				if( pObj->HasValue( pVar ) )
				{
					int Dimension = 1;

					// Type of variable to string
					int Type = pVar->GetType();
					
					// is it an array?
					int ArrayFlag = Type & VARTYPEARRAYMASK;

					// mask pure scalar type
					Type &= ~VARTYPEARRAYMASK;

					// now dispatch for all know types
					for( int i = 0; i < IStoredVar::NumberOfScalarTypes; i++ )
					{

						// compare typename
						if( Type == TypeStringLookup[i].Type )
						{
							File << TypeStringLookup[i].Name << ":";

							// leave the loop
							break;
						}
					}



					// append name of variable
					File << pVar->Name();
					
					

					//
					// is it an array?
					//
					if( ArrayFlag )
					{

						//
						// we are going to write everything out as double
						// (there are no string arrays, currently
						//
						double FixedBuffer[_FIXEDBUFFERSIZE];
						double *pBuffer = FixedBuffer;

						// set Dimension to 0 to avoid copying of data to the NULL pointer in GetValue()
						Dimension = 0;
						
						// retrieve Dimension
						pObj->GetValue( pVar, pBuffer, Dimension );


						// write dimension and add a whitespace
						File << "[" << Dimension << "]";


						//
						// get array data 
						//

						//
						// if there is too much data for the array on the stack
						// allocate a new array 
						//
						if( Dimension > _FIXEDBUFFERSIZE )
							pBuffer = new double( Dimension );
						

						// fetch array
						pObj->GetValue( pVar, pBuffer, Dimension );


						//
						// write values
						//
						for( i = 0; i < Dimension; ++i )
							File << " " << pBuffer[ i ] ;

						//
						// if data has been allocated, delete it again
						//
						if( Dimension > _FIXEDBUFFERSIZE )
							delete[] pBuffer;
										
					}
					else
					{
						//
						// not an array
						//
						
						//
						// retrieve data as a string
						//

						string Data;

						pObj->GetValue( pVar, Data );

						File << " " << Data ;
					}

					// add a linefeed for the next variable
					File << endl;
				}


			} STOREDTYPEENDFOR

			// add a blank line between objects 
			File << endl;

			// objects from the iterator have to be released again
			pObj->Release();

		} STOREDTYPEENDFOR

	} STOREDTYPEENDFOR


	return true;
}