// CStorageMemory.h
//**********************************************************************************************
//
//
//				(C) Copyright 1998 by Dipl. Phys. Joerg Plewe, HARDCODE Development
//					All rights reserved. Copying, modification,
//					distribution or publication without the prior written 
//					consent of the author is prohibited.
//
//
//	Filename:		CStorageMemory.h
//	Author:			Herkules
//  Created:		9810xx 
//
//	Module:			StorageMemory
//  Dependecies:	STL, create a DLL
//
//  History:
//  $Date$		$Who$		$What$
//	19981220	Herkules	GetValue() now receives a <type>*& instead of the pointer value
//							If the pointer is NULL, storage tries to set the pointer to address
//							of the internal data
//	19981228	Herkules	BUG: When Flush()ing objects from a derived driver (e.g. Library) with
//							_KeepData==false more than once, it re-Compiles() an already cleared object!
//							Fix by introduction of the m_bIsFlushed flag. 
//	19990203	Herkules	Introduced CStoredObject::ReAllocate()
//
/*	Purpose:

These are the memory 'driver' classes for the storage engine.
Classes are derived from the 'high-level' classes in CStorage.h

*/
//**********************************************************************************************

#pragma once


#ifndef _H_CStorageMemory
#define _H_CStorageMemory

#include "../CStorage.h" 





//
// forwards
//
class CStorageMemory;
class IStoredVarMemory;
class IUncompiledValue;

template <class T> bool AddTypedValue( T _Value, int _Size );




class IUncompiledValue
{
public:
	virtual ~IUncompiledValue()	{}

	virtual int CompiledSize() = 0;
	virtual unsigned char* Compile( unsigned char* _pData ) = 0;

	virtual CStoredVar* GetVar()	{ return m_pVar; }

protected:
	CStoredVar* m_pVar;

};



//----------------------------------------------------------------------------------------------------------
// the StoredClass for the memory database
//

class CStoredObjectMemory : public CStoredObject
{
	typedef CStoredObject super;
	
	friend class CStoredClassMemory;
public:

	
	virtual bool HasValue( IStoredVar* _pVar );

	
	
	// add data
	virtual bool AddValue( IStoredVar* _pVar, unsigned char _Value );
	virtual bool AddValue( IStoredVar* _pVar, short _Value );
	virtual bool AddValue( IStoredVar* _pVar, int _Value );
	virtual bool AddValue( IStoredVar* _pVar, LongInt _Value );
	virtual bool AddValue( IStoredVar* _pVar, float _Value );
	virtual bool AddValue( IStoredVar* _pVar, double _Value );
	virtual bool AddValue( IStoredVar* _pVar, const string& _Value, int _MaxSize = 0 );


	// array 
	virtual bool AddValue( IStoredVar* _pVar, const unsigned char* _pValue, int _Size = 1 );
	virtual bool AddValue( IStoredVar* _pVar, const short* _pValue, int _Size = 1);
	virtual bool AddValue( IStoredVar* _pVar, const int* _pValue, int _Size = 1);
	virtual bool AddValue( IStoredVar* _pVar, const LongInt* _pValue, int _Size = 1 );
	virtual bool AddValue( IStoredVar* _pVar, const float* _pValue, int _Size = 1);
	virtual bool AddValue( IStoredVar* _pVar, const double* _pValue, int _Size = 1);


	// retrieve data
	virtual bool GetValue( IStoredVar* _pVar, unsigned char& _Value );
	virtual bool GetValue( IStoredVar* _pVar, short& _Value );
	virtual bool GetValue( IStoredVar* _pVar, int& _Value );
	virtual bool GetValue( IStoredVar* _pVar, LongInt& _Value );
	virtual bool GetValue( IStoredVar* _pVar, float& _Value );
	virtual bool GetValue( IStoredVar* _pVar, double& _Value );
	virtual bool GetValue( IStoredVar* _pVar, string& _Value );

	// array 
	virtual bool GetValue( IStoredVar* _pVar, unsigned char*& _pValue, int& _Size );
	virtual bool GetValue( IStoredVar* _pVar, short*& _pValue, int& _Size );
	virtual bool GetValue( IStoredVar* _pVar, int*& _pValue, int& _Size );
	virtual bool GetValue( IStoredVar* _pVar, LongInt*& _pValue, int& _Size );
	virtual bool GetValue( IStoredVar* _pVar, float*& _pValue, int& _Size );
	virtual bool GetValue( IStoredVar* _pVar, double*& _pValue, int& _Size );



	// modify data
	virtual bool SetValue( IStoredVar* _pVar, unsigned char _Value );
	virtual bool SetValue( IStoredVar* _pVar, short _Value );
	virtual bool SetValue( IStoredVar* _pVar, int _Value );
	virtual bool SetValue( IStoredVar* _pVar, LongInt _Value );
	virtual bool SetValue( IStoredVar* _pVar, float _Value );
	virtual bool SetValue( IStoredVar* _pVar, double _Value );
	virtual bool SetValue( IStoredVar* _pVar, const string& _Value );


	// array 
	virtual bool SetValue( IStoredVar* _pVar, unsigned char* _pValue, int& _Size );
	virtual bool SetValue( IStoredVar* _pVar, short* _pValue, int& _Size );
	virtual bool SetValue( IStoredVar* _pVar, int* _pValue, int& _Size );
	virtual bool SetValue( IStoredVar* _pVar, LongInt* _pValue, int& _Size );
	virtual bool SetValue( IStoredVar* _pVar, float* _pValue, int& _Size );
	virtual bool SetValue( IStoredVar* _pVar, double* _pValue, int& _Size );



	// references to other objects
	virtual bool GetValue( IStoredVar* _pVar, IStoredObject** _pObj, int& _Size ); 
	virtual bool AddValue( IStoredVar* _pVar, IStoredObject** _pObj, int _Size = 1 ); 
	virtual bool SetValue( IStoredVar* _pVar, IStoredObject** _pObj, int& _Size ); 


	//
	// memory objects can neither be flushed not released,
	// bc. they ARE the data of the storage
	//
	virtual bool Flush( bool _KeepData = false )	
	{ 

		// we try to flush a second time without data change.
		if( m_bIsFlushed ) return false;

		// in any case, the object is flushed NOW
		m_bIsFlushed = true;

		// if already compiled, theres nothing to do on Flush()
		if( m_pCompiledObject ) return true;

		// we can override the flag setting from above again, if Compile() fails
		m_bIsFlushed = Compile(); 

		return m_bIsFlushed;

	}

	virtual bool Refresh()						
	{
		
		// When it is refreshed, it is clean and can be counted as flushed 
		// But ONLY, if there is any data! While constructing the object, before
		// first Flush(), Refresh() should not do anything
		if( m_ObjectSize > 0 ) 
			m_bIsFlushed = true;

		return super::Refresh(); 
	}


	// throw away all existing data and start from scratch 
	virtual bool ReAllocate();


	void AddUncompiledValue( IUncompiledValue* _pValue )
	{
		m_lUncompiledValues.push_back( _pValue );
	}


	// size of the structure used for references
	// this is defined as a virtual function enabling derived classes to use
	// our Compile() method
	virtual int SizeOfReference();
	virtual unsigned char* CompileReference( unsigned char* _pData, IStoredObject* _pObj );
	virtual IStoredObject* CreateFromReferenceData( void* _pRefDat );


	// retrieve information about the compiled data
	OBJECTSIZE FlushedObjectSize();	
	unsigned char* FlushedObjectData()			{ return m_pCompiledObject; }	


protected:

	CStoredObjectMemory( int _ObjectIndex, IStoredClass* _pStoredClass ) : 
	CStoredObject( _ObjectIndex, _pStoredClass ),
	m_pCompiledObject( NULL ),
	m_bIsFlushed( false )
	{}

	virtual ~CStoredObjectMemory();



	//
	// services for derived classes
	//


	// (de)allocation of memory for the compiled object is done with
	// virtual functions, so that derived class can overload them
	virtual unsigned char* AllocateCompiledObject( int _Size )	
	{ 
		return new unsigned char[ _Size ]; 
	}
	
	// deallocate 
	virtual void DeAllocateCompiledObject()				
	{ 
		if( m_pCompiledObject )
			delete[] m_pCompiledObject;
		m_pCompiledObject = NULL;
	}



	// use external data
	void AssignData( unsigned char* pData, OBJECTSIZE _Size );


	// Place the pointer to the memory where is object is compiled here!
	unsigned char* m_pCompiledObject;

	// size of object after Compile()
	OBJECTSIZE m_ObjectSize;


	// tells wether the object is flushed or not 
	// this is complicated an VERY subtile!
	// an object is counted as 'flushed', when its clean -
	// means data has been compiled/written out and nobody else
	// touched it afterwards.
	// This is used to handle multiple flushes on the same object.
	// Set to false by constructor and SetValue()
	// Set to true by Compile();
	bool m_bIsFlushed;


private:

	//
	// pack the object to a single, fixed-sized memory block
	//
	bool Compile();


	//
	// this function no longer is a member, but locally declared in CStorageMemory.cpp
	//
	// static char* FindCompiledValue( CStoredVar* _pVar, char* _pDataBuffer );


	//
	// get a value as a string
	//
	bool GetCompiledStringValue( CStoredVar* _pVar, string& _String );


	// add as value as a string
	friend bool AddStringValue( CStoredObjectMemory& _Object, CStoredVar* _pVar, const string& _String, int _MaxSize );

	
	// store the values of a still uncompiled object in a list
	std::list<IUncompiledValue*> m_lUncompiledValues;

}; // CStoredObjectMemory




// ----------------------------------------------------------------------------
//
// class for object access in an array manner
//
class CStoredObjectArrayMemory : public CStoredObjectArray
{
	friend class CStoredClassMemory;
public:

	// number of objects
	virtual int	Count()				{ return m_Count; }


	//
	// access a certain object by index
	//
	virtual IStoredObject* Get( int _Index, bool _WithData )	
	{ 

		// the _WithData flag is ignored - memory database cannot be without data
		
		// 
		// check wether index is in range
		// if it is ok, return the pointer from the array
		//
		if( _Index >= 0  &&  _Index < m_Count )
			return m_raObjects[ _Index ];


		//
		// _Index out of range
		//
		return NULL;
	}


protected:

	CStoredObjectArrayMemory( const std::vector<CStoredObjectMemory*>& _aObjects, const int _Count ) :
	m_raObjects( _aObjects ),
	m_Count( _Count )
	{
	}


private:

	//
	// Objects are organizes in a std::vector
	// we take a reference to that vector
	//
	const std::vector<CStoredObjectMemory*>& m_raObjects;

	// number of objects in the vector
	int m_Count;

};








// ----------------------------------------------------------------------------
//
// class for object access in an array manner
//
class CStoredObjectListMemory : public CStoredObjectList
{
	friend class CStoredClassMemory;
public:

	//
	// number of objects
	//
	virtual int	Count()				{ return m_Count; }

	//
	// retrieve pointer to first object
	//
	virtual IStoredObject*	First( bool _WithData ) 
	{

		// the _WithData flag is ignored - memory database cannot be without data
		
		if( m_Count != 0 )
			return m_raObjects[ 0 ];

		// vector is empty
		return NULL;
	}


	//
	// pointer to next object
	//
	virtual IStoredObject* Next( bool _WithData )
	{
		// the _WithData flag is ignored - memory database cannot be without data
		
		if( (++m_CurrentIndex) < m_Count )
			return m_raObjects[ m_CurrentIndex ];

		// end of vector reached
		return NULL;
	}


protected:
	CStoredObjectListMemory( const std::vector<CStoredObjectMemory*>& _aObjects, const int _Count ) :
		m_raObjects( const_cast<std::vector<CStoredObjectMemory*>&>(_aObjects) ),
		m_Count( _Count ),
		m_CurrentIndex( 0 )
	{
	}


private:
	
	// Object are organize in a std::vector
	// we take a reference to that vector
	std::vector<CStoredObjectMemory*>& m_raObjects;

	// number of objects in the vector
	int m_Count;

	// Current index in the array
	int m_CurrentIndex;
};






// ---------------------------------------------------------------------------------------------------
//
// CStoredClass for the memory database
//
class CStoredClassMemory : public CStoredClass
{

	typedef CStoredClass super;

public:
	// Constuctor
	//
	// CStoredClassMemory does not rely on _pStorage being
	// a CStorageMemory. It can wor with any class derive from CStorage
	//
	CStoredClassMemory( const string& _Classname, 
						const STOREDCLASS_ID _ClassID, 
						CStorage* _pStorage, 
						IStoredClass* const _pBaseClass = NULL );
	virtual ~CStoredClassMemory();



	//
	// add data
	//
	virtual IStoredObject* AddObject( );


	//
	// get access objects
	//
	virtual IStoredObjectArray* GetObjectArray()		{ return new CStoredObjectArrayMemory( m_aObjects, m_NumObjects ); }
	virtual IStoredObjectList* GetObjectList() 			{ return new CStoredObjectListMemory( m_aObjects, m_NumObjects ); }



protected:

	// Take an object and add it to the classes list of objects
	bool AssignObject( IStoredObject* _pObject );


	// the CreateVar method changes the repository. 
	virtual CStoredVar* CreateVar( const string& _Varname, STOREDVAR_TYPE _Type, STOREDVAR_ID _ID );

private:

	// counting objects
	int m_NumObjects;

	// objects are maintained in an array/vector
	std::vector<CStoredObjectMemory*> m_aObjects;

}; // CStoredClassMemory








// ----------------------------------------------------------------------------
// 
//
// CStorage for the memory database
//
//
class CStorageMemory : public CStorage
{

	typedef CStorage super;

public:

	// constructor
	CStorageMemory( IStorageEvent* _pEvent = NULL ) : super( _pEvent )
	{}


protected:
	//
	//
	//
	virtual CStoredClass* CreateClass( const string& _Classname, const STOREDCLASS_ID _ClassID, IStoredClass* const _pBaseClass = NULL )	
	{ 
		return new CStoredClassMemory( _Classname, _ClassID, this, _pBaseClass ); 
	}

	virtual ~CStorageMemory() {}

};





#endif // _H_CStorageMemory
