// main.cpp
//**********************************************************************************************
//
//
//				(C) Copyright 1998 by Dipl. Phys. Joerg Plewe, HARDCODE Development
//					All rights reserved. Copying, modification,
//					distribution or publication without the prior written 
//					consent of the author is prohibited.
//
//
//	Filename:		main.cpp
//	Author:			Herkules
//  Created:		19990207
//  Version:		$Version$
//
//	Module:			StorageExamples
//  Dependecies:	$Dependencies$
//
//  History:
//  $Date$		$Who$		$What$
//
/*	Purpose:

Setup a simple database and fill in some values


*/
//**********************************************************************************************


#include <iostream>
#include <assert.h>

#include "..\..\Storage\IStorage.h"




//
// define some names for classes
//
#define NAME_CLASS_FOO "FooClass"


//
// define some names for variables
//
#define NAME_VAR_INT1	"Integer1"
#define TYPE_VAR_INT1	IStoredVar::Int

#define NAME_VAR_STR1	"String1"
#define TYPE_VAR_STR1	IStoredVar::String

#define NAME_VAR_REF1	"Reference1"
#define TYPE_VAR_REF1	IStoredVar::ReferenceArray



void main()
{

	//
	// locals 
	//	
	bool ret;

	// we will need an engine
	IStorageEngine* pEngine;

	// we do it with just one class
	IStoredClass*	pClassFoo;

	// a couple of variables
	IStoredVar*		pVarInteger1;
	IStoredVar*		pVarString1;
	IStoredVar*		pVarReference1;

	// some objects
	IStoredObject*	pFooObj1;
	IStoredObject*	RefArray[100];


	// create the libary
	pEngine = IStorageEngine::New( IStorageEngine::Memory );
	
	// connect: does nothing for a memory database
	ret = pEngine->Connect( "" );


	// add a class to repository
	pClassFoo = pEngine->AddClass( NAME_CLASS_FOO );
	assert( pClassFoo );

	
	//
	// add variables to the class
	//
	pVarInteger1	= pClassFoo->AddVar( NAME_VAR_INT1, TYPE_VAR_INT1 );
	pVarString1		= pClassFoo->AddVar( NAME_VAR_STR1, TYPE_VAR_STR1 );
	pVarReference1	= pClassFoo->AddVar( NAME_VAR_REF1, TYPE_VAR_REF1 );


	// create the first object
	pFooObj1 = pClassFoo->AddObject();
	assert( pFooObj1 );


	//---------------------------------------------------------------------------------------------
	//
	// now insert some values into the object
	//


	// Fill the integer variable. The data comes from a string
	ret = pFooObj1->AddValue( pVarInteger1, "123" );  

	// Fill the string variable. The data comes from a float :-)
	ret = pFooObj1->AddValue( pVarString1, 456.0f );  


	// Prepate FooObj to reference five other objects
	RefArray[0] = NULL;
	RefArray[1] = NULL;
	RefArray[2] = NULL;
	RefArray[3] = NULL;
	RefArray[4] = NULL;
	ret = pFooObj1->AddValue( pVarReference1, RefArray, 5 );  
	

	// flush the object, make it ready for use
	ret = pFooObj1->Flush();


	// Read a single integer
	int Integer1;
	ret = pFooObj1->GetValue( pVarInteger1, Integer1 );

	// Read the same variable into a float
	float Float1;
	ret = pFooObj1->GetValue( pVarInteger1, Float1 );

	// Read the same variable into a string
	string String1;
	ret = pFooObj1->GetValue( pVarInteger1, String1 );

	// Get the string variable as a string
	ret = pFooObj1->GetValue( pVarString1, String1 );


	// 
	// shutdown the library
	//
	pEngine->Close();
	pEngine->Release();


}