// main.cpp
//**********************************************************************************************
//
//
//				(C) Copyright 1998 by Dipl. Phys. Joerg Plewe, HARDCODE Development
//					All rights reserved. Copying, modification,
//					distribution or publication without the prior written 
//					consent of the author is prohibited.
//
//
//	Filename:		main.cpp
//	Author:			Herkules
//  Created:		19990207
//  Version:		$Version$
//
//	Module:			StorageExamples
//  Dependecies:	$Dependencies$
//
//  History:
//  $Date$		$Who$		$What$
//
/*	Purpose:

Setup a simple database and fill in some values


*/
//**********************************************************************************************


#include <iostream>
#include <iomanip>
#include <assert.h>
#include <time.h>
#include <conio.h>


#include "..\..\Storage\IStorage.h"




//
// define some names for classes
//
#define NAME_CLASS_FOO "MassDataClass"


//
// define some names for variables
//
#define NAME_VAR_INT1	"Integer"
#define TYPE_VAR_INT1	IStoredVar::IntArray

#define NAME_VAR_STR1	"String"
#define TYPE_VAR_STR1	IStoredVar::String




void main()
{

	//
	// locals 
	//	
	bool ret;

	// we will need an engine
	IStorageEngine* pEngine;

	// we do it with just one class
	IStoredClass*	pClassFoo;

	// a couple of variables
	IStoredVar*		pVarIntegerArray;
	IStoredVar*		pVarString;

	// some objects
	IStoredObject*	pFooObj;


	cout << "Creating file library: MassData.hsl" << endl;



	// create the libary
	pEngine = IStorageEngine::New( IStorageEngine::Library );
	
	// connect: does nothing for a memory database
	ret = pEngine->Connect( "MassData.hsl" );


	//
	// Create or connect to repository
	//

	pClassFoo = pEngine->GetClass( NAME_CLASS_FOO );
	if( pClassFoo )
	{
		cout << "IStoredClass MassDataClass seems to exist in the lib." << endl;
		cout << "Trying to connect to existing repository....";
	}
	else
	{
		cout << "Creating a new repository....";

		pClassFoo = pEngine->AddClass( NAME_CLASS_FOO );
	}

	
	//
	// add (or get ) variables to the class
	//
	pVarIntegerArray	= pClassFoo->AddVar( NAME_VAR_INT1, TYPE_VAR_INT1 );
	pVarString			= pClassFoo->AddVar( NAME_VAR_STR1, TYPE_VAR_STR1 );

	if( pClassFoo && pVarIntegerArray && pVarString )
	{
		cout << "... succeeded!" << endl;
	}
	else
	{
		cout << "... failed!" << endl;
		exit( 0 );
	}
  

	
	//
	// now build the database
	//

	cout << "Adding 10000 object to the database (one . for 100 objects)" << endl;
	cout << "This will add approx. 10MB to the file." << endl;

	time_t   start, finish;
	double   elapsed_time;
	time( &start );

	
	int Array[256];		// 1kb

	for( int i = 0; i < 10000; ++i )
	{
		// Add a new object 	
		pFooObj = pClassFoo->AddObject();
		assert( pFooObj );

		
		// Provide data for the object
		Array[0] = i;
		ret = pFooObj->AddValue( pVarIntegerArray, Array, 256 );

		// Add the index as a string
		ret = pFooObj->AddValue( pVarString, i );

		// Flush and release the object
		pFooObj->Flush();
		pFooObj->Release();


		if( (i % 100) == 0 )
			cout << ".";

	}
		

	cout << " done" << endl;

	time( &finish );   
	elapsed_time = difftime( finish, start );
	
	cout << "Insertion took " << showpoint << elapsed_time << " seconds" << endl;



	//---------------------------------------------------------------------------------------------
	//
	// now the database is filled and we are going to read it sequentially
	//


	cout << endl << "Now reading in the data sequentially" << endl;
	
	IStoredObjectList* pList;


	pList = pClassFoo->GetObjectList();
	assert( pList );

	int Count = pList->Count();
	cout << "Found " << Count << " objects in the library" << endl;
	

	//
	//
	//

	cout << "Reading all object (one . for 100 objects)" << endl;


	time( &start );

	pFooObj = pList->First();
	Count = 0;


	string String;
	int Integer;

	while( pFooObj )
	{


		ret = pFooObj->GetValue( pVarIntegerArray, Integer );
		ret = pFooObj->GetValue( pVarString, String );
	
		
		pFooObj->Release();
		pFooObj = pList->Next();

		if( (Count++ % 100) == 0 )
			cout << ".";
	}

	cout << " done" << endl;

	time( &finish );   
	elapsed_time = difftime( finish, start );
	
	cout << "Read took " << showpoint << elapsed_time << " seconds" << endl;


	cout << endl << "Press any key....";



	// 
	// shutdown the library
	//
	pEngine->Close();
	pEngine->Release();


	int c = _getch();
}